from math import floor

from sklearn.cluster import KMeans as sk_KMeans

from astartes.samplers import AbstractSampler


class KMeans(AbstractSampler):
    def _sample(self):
        """Implements the K-Means sampler to identify clusters."""
        # use the sklearn kmeans model
        default_kwargs = {
            "n_clusters": self.get_config("n_clusters", floor(len(self.X) * 0.1) + 1),
            "init": self.get_config("init", "k-means++"),
            "n_init": self.get_config("n_init", 1),
            "max_iter": self.get_config("max_iter", 300),
            "tol": self.get_config("tol", 1e-4),
            "verbose": self.get_config("verbose", 0),
            "random_state": self.get_config("random_state", None),
            "copy_x": self.get_config("copy_x", True),
        }
        if self.get_config("algorithm", False):
            default_kwargs["algorithm"] = self.get_config("algorithm")

        kmeanModel = sk_KMeans(**default_kwargs).fit(self.X)
        self._samples_clusters = kmeanModel.labels_



# convenience import to enable 'from astartes import train_test_split'
from .main import train_test_split, train_val_test_split

__version__ = "1.2.2"

# DO NOT do this:
# from .molecules import train_test_split_molecules
# on installations without the optional molecules install, this would cause
# an exception



"""
Implements the Sample set Partitioning based on join X-Y distances
algorithm as originally described by Saldanha and coworkers in
"A method for calibration and validation subset partitioning"
doi:10.1016/j.talanta.2005.03.025

This implementation has been validated against their original source
code implementation, which can be found in the paper linked above.
The corresponding unit tests reflect the expected output from
the original implemenation. The breaking of ties is different
compared to the original, but this is ultimately a minor and
likely inconsequential difference.
"""
import numpy as np
from scipy.spatial.distance import pdist, squareform

from astartes.samplers import AbstractSampler
from astartes.utils.exceptions import InvalidConfigurationError
from astartes.utils.fast_kennard_stone import fast_kennard_stone


class SPXY(AbstractSampler):
    def _before_sample(self):
        if self.y is None:
            raise InvalidConfigurationError("SPXY sampler requires both X and y arrays. Provide y or switch to kennard_stone.")

    def _sample(self):
        """
        Implements the SPXY algorithm as described by Saldahna et al.
        """
        distance_metric = self.get_config("metric", "euclidean")

        y_2d = self.y[:, np.newaxis]

        y_pdist = pdist(y_2d, metric=distance_metric)
        y_pdist = np.divide(y_pdist, np.amax(y_pdist))

        X_dist = pdist(self.X, metric=distance_metric)
        X_dist = np.divide(X_dist, np.amax(X_dist))

        # sum the distances as per eq. 3 of Saldahna, set diagonal to nan
        spxy_distance = squareform(y_pdist + X_dist)

        self._samples_idxs = fast_kennard_stone(spxy_distance)



import sklearn
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
from tabulate import tabulate

import astartes


def generate_regression_results_dict(
    sklearn_model,
    X,
    y,
    samplers=["random"],
    random_state=0,
    samplers_hopts={},
    train_size=0.8,
    val_size=0.1,
    test_size=0.1,
    print_results=False,
    additional_metrics={},
):
    """
    Helper function to train a sklearn model using the provided data
    and provided sampler types.

    Args:
        X (np.array, pd.DataFrame): Numpy array or pandas DataFrame of feature vectors.
        y (np.array, pd.Series): Targets corresponding to X, must be of same size.
        train_size (float, optional): Fraction of dataset to use in training set. Defaults to 0.8.
        val_size (float, optional): Fraction of dataset to use in validation set. Defaults to 0.1.
        test_size (float, optional): Fraction of dataset to use in test set. Defaults to 0.1.
        random_state (int, optional): The random seed used throughout astartes.
        samplers_hopts (dict, optional): Should be a dictionary of dictionaries with the keys specifying
                                         the sampler and the values being another dictionary with the
                                         corresponding hyperparameters. Defaults to {}.
        print_results (bool, optional): whether to print the resulting dictionary as a neat table
        additional_metrics (dict, optional): mapping of name (str) to metric (func) for additional metrics
                                             such as those in sklearn.metrics or user-provided functions

    Returns:
        dict: nested dictionary with the format of
            {
                sampler: {
                    'mae':{
                        'train': [],
                        'val': [],
                        'test': [],
                    },
                    'rmse':{
                        'train': [],
                        'val': [],
                        'test': [],
                    },
                    'R2':{
                        'train': [],
                        'val': [],
                        'test': [],
                    },
                },
            }
    """
    if not isinstance(sklearn_model, sklearn.base.BaseEstimator):
        raise astartes.utils.exceptions.InvalidModelTypeError(
            "Model must be an sklearn model"
        )

    final_dict = {}
    for sampler in samplers:
        error_dict = {
            "mae": {
                "train": [],
                "val": [],
                "test": [],
            },
            "rmse": {
                "train": [],
                "val": [],
                "test": [],
            },
            "R2": {
                "train": [],
                "val": [],
                "test": [],
            },
        }

        # obtain indices
        (
            _,
            _,
            _,
            train_indices,
            val_indices,
            test_indices,
        ) = astartes.train_val_test_split(
            X,
            train_size=train_size,
            val_size=val_size,
            test_size=test_size,
            sampler=sampler,
            random_state=random_state,
            hopts=samplers_hopts.get(sampler, dict()),
            return_indices=True,
        )

        # create data splits
        X_train = X[train_indices]
        X_val = X[val_indices]
        X_test = X[test_indices]

        y_train = y[train_indices]
        y_val = y[val_indices]
        y_test = y[test_indices]

        # fit the model to the training data
        sklearn_model.fit(X_train, y_train)

        # get predictions
        y_pred_train = sklearn_model.predict(X_train)
        y_pred_val = sklearn_model.predict(X_val)
        y_pred_test = sklearn_model.predict(X_test)

        # store MAEs
        train_mae = mean_absolute_error(y_train, y_pred_train)
        error_dict["mae"]["train"].append(train_mae)

        val_mae = mean_absolute_error(y_val, y_pred_val)
        error_dict["mae"]["val"].append(val_mae)

        test_mae = mean_absolute_error(y_test, y_pred_test)
        error_dict["mae"]["test"].append(test_mae)

        # store RMSEs
        train_rmse = mean_squared_error(y_train, y_pred_train, squared=False)
        error_dict["rmse"]["train"].append(train_rmse)

        val_rmse = mean_squared_error(y_val, y_pred_val, squared=False)
        error_dict["rmse"]["val"].append(val_rmse)

        test_rmse = mean_squared_error(y_test, y_pred_test, squared=False)
        error_dict["rmse"]["test"].append(test_rmse)

        # store R2
        train_R2 = r2_score(y_train, y_pred_train)
        error_dict["R2"]["train"].append(train_R2)

        val_R2 = r2_score(y_val, y_pred_val)
        error_dict["R2"]["val"].append(val_R2)

        test_R2 = r2_score(y_test, y_pred_test)
        error_dict["R2"]["test"].append(test_R2)

        final_dict[sampler] = error_dict

        for metric_name, metric_function in additional_metrics.items():
            error_dict[metric_name]["train"] = metric_function(y_train, y_pred_train)
            error_dict[metric_name]["val"] = metric_function(y_val, y_pred_val)
            error_dict[metric_name]["test"] = metric_function(y_test, y_pred_test)

        if print_results:
            print(f"\nDisplaying results for {sampler} sampler")
            display_results_as_table(error_dict)

    return final_dict


def display_results_as_table(error_dict):
    """Helper function to print a dictionary as a neat tabulate"""
    headers = ["Train", "Val", "Test"]
    table = []
    for key, val in error_dict.items():
        table_tmp = [key.upper()]
        table_tmp.extend([val[0] for val in val.values()])
        table.append(table_tmp)
    print(tabulate(table, headers=headers))



import numpy as np

# at this point we have successfully verified that rdkit is installed, so we can do this:
from rdkit.rdBase import SeedRandomNumberGenerator

from astartes import train_test_split, train_val_test_split
from astartes.main import DEFAULT_RANDOM_STATE
from astartes.utils.aimsim_featurizer import featurize_molecules

SeedRandomNumberGenerator(DEFAULT_RANDOM_STATE)


def train_val_test_split_molecules(
    molecules: np.array,
    y: np.array = None,
    labels: np.array = None,
    train_size: float = 0.8,
    val_size: float = 0.1,
    test_size: float = 0.1,
    sampler: str = "random",
    random_state: int = None,
    hopts: dict = {},
    fingerprint: str = "morgan_fingerprint",
    fprints_hopts: dict = {},
    return_indices: bool = False,
):
    """Deterministic train_test_splitting of molecules (SMILES strings or RDKit objects).

    Args:
        molecules (np.array): List of SMILES strings or RDKit molecule objects representing molecules or reactions.
        y (np.array, optional): Targets corresponding to SMILES, must be of same size. Defaults to None.
        labels (np.array, optional): Labels corresponding to SMILES, must be of same size. Defaults to None.
        train_size (float, optional): Fraction of dataset to use in training set. Defaults to 0.8.
        val_size (float, optional): Fraction of dataset to use in validation set. Defaults to 0.1.
        test_size (float, optional): Fraction of dataset to use in test set. Defaults to 0.1.
        sampler (str, optional): Sampler to use, see IMPLEMENTED_INTER/EXTRAPOLATION_SAMPLERS. Defaults to "random".
        random_state (int, optional): The random seed used throughout astartes. Defaults to 42.
        hopts (dict, optional): Hyperparameters for the sampler used above. Defaults to {}.
        fingerprint (str, optional): Molecular fingerprint to be used from AIMSim. Defaults to "morgan_fingerprint".
        fprints_hopts (dict, optional): Hyperparameters for AIMSim featurization. Defaults to {}.
        return_indices (bool, optional): True to return indices of train/test after the values. Defaults to False.

    Returns:
        np.array: X, y, and labels train/val/test data, or indices.
    """
    if sampler == "scaffold":
        X = molecules
    else:
        X = featurize_molecules(molecules, fingerprint, fprints_hopts)
    return train_val_test_split(
        X,
        y=y,
        labels=labels,
        test_size=test_size,
        val_size=val_size,
        train_size=train_size,
        sampler=sampler,
        random_state=random_state,
        hopts=hopts,
        return_indices=return_indices,
    )


def train_test_split_molecules(
    molecules: np.array,
    y: np.array = None,
    labels: np.array = None,
    train_size: float = 0.75,
    test_size: float = None,
    sampler: str = "random",
    random_state: int = None,
    hopts: dict = {},
    fingerprint: str = "morgan_fingerprint",
    fprints_hopts: dict = {},
    return_indices: bool = False,
):
    """Deterministic train/test splitting of molecules (SMILES strings or RDKit objects).

    Args:
        molecules (np.array): List of SMILES strings or RDKit molecule objects representing molecules or reactions.
        y (np.array, optional): Targets corresponding to SMILES, must be of same size. Defaults to None.
        labels (np.array, optional): Labels corresponding to SMILES, must be of same size. Defaults to None.
        train_size (float, optional): Fraction of dataset to use in training (test+train~1). Defaults to 0.75.
        test_size (float, optional): Fraction of dataset to use in test set. Defaults to None.
        sampler (str, optional): Sampler to use, see IMPLEMENTED_INTER/EXTRAPOLATION_SAMPLERS. Defaults to "random".
        random_state (int, optional): The random seed used throughout astartes. Defaults to None.
        hopts (dict, optional): Hyperparameters for the sampler used above. Defaults to {}.
        fingerprint (str, optional): Molecular fingerprint to be used from AIMSim. Defaults to "morgan_fingerprint".
        fprints_hopts (dict, optional): Hyperparameters for AIMSim featurization. Defaults to {}.
        return_indices (bool, optional): True to return indices of train/test after the values. Defaults to False.

    Returns:
        np.array: X, y, and labels train/test data, or indices.
    """
    # turn the smiles into an input X
    if sampler == "scaffold":
        X = molecules
    else:
        X = featurize_molecules(molecules, fingerprint, fprints_hopts)

    # call train test split with this input
    return train_test_split(
        X,
        y=y,
        labels=labels,
        test_size=test_size,
        train_size=train_size,
        sampler=sampler,
        random_state=random_state,
        hopts=hopts,
        return_indices=return_indices,
    )



from difflib import get_close_matches

from astartes.samplers import (
    ALL_SAMPLERS,
    DBSCAN,
    SPXY,
    KennardStone,
    KMeans,
    MolecularWeight,
    OptiSim,
    Random,
    Scaffold,
    SphereExclusion,
    TargetProperty,
    TimeBased,
)
from astartes.utils.exceptions import SamplerNotImplementedError


class SamplerFactory:
    def __init__(self, sampler):
        """Initialize SamplerFactory and copy a lowercased 'sampler' into an attribute.

        Args:
            sampler (string): The desired sampler.
        """
        self.sampler = sampler.lower()

    def get_sampler(self, X, y, labels, hopts):
        """Instantiate (which also performs fitting) and return the sampler.

        Args:
            X (np.array): Feature array.
            y (np.array): Target array.
            labels (np.array): Label array.
            hopts (dict): Hyperparameters for the sampler.

        Raises:
            SamplerNotImplementedError: Raised when an non-existent or not yet implemented sampler is requested.

        Returns:
            astartes.sampler: The fit sampler instance.
        """
        if self.sampler == "random":
            sampler_class = Random
        elif self.sampler == "kennard_stone":
            sampler_class = KennardStone
        elif self.sampler == "dbscan":
            sampler_class = DBSCAN
        elif self.sampler == "kmeans":
            sampler_class = KMeans
        elif self.sampler == "sphere_exclusion":
            sampler_class = SphereExclusion
        elif self.sampler == "molecular_weight":
            sampler_class = MolecularWeight
        elif self.sampler == "optisim":
            sampler_class = OptiSim
        elif self.sampler == "spxy":
            sampler_class = SPXY
        elif self.sampler == "scaffold":
            sampler_class = Scaffold
        elif self.sampler == "time_based":
            sampler_class = TimeBased
        elif self.sampler == "target_property":
            sampler_class = TargetProperty
        else:
            possiblity = get_close_matches(
                self.sampler,
                ALL_SAMPLERS,
                n=1,
            )
            addendum = (
                " Did you mean '{:s}'?".format(possiblity[0])
                if possiblity
                else " Try help(train_test_split)."
            )
            raise SamplerNotImplementedError(
                "Sampler {:s} has not been implemented.".format(self.sampler) + addendum
            )
        return sampler_class(X, y, labels, hopts)



"""
The Sphere Exclusion clustering algorithm.

This re-implementation draws from this blog post on the RDKit blog,
though abstracted to work for arbitrary feature vectors:
http://rdkit.blogspot.com/2020/11/sphere-exclusion-clustering-with-rdkit.html
As well as this paper:
https://www.daylight.com/cheminformatics/whitepapers/ClusteringWhitePaper.pdf

But instead of using tanimoto similarity, which has a domain between zero and
one, it uses euclidian distance to enable processing arbitrary valued
vectors.
"""
from math import floor

import numpy as np
from scipy.spatial.distance import pdist, squareform

from astartes.samplers import AbstractSampler


class SphereExclusion(AbstractSampler):
    def _sample(self):
        """Cluster X according to a Sphere Exclusion-like algorithm with arbitrary distance metrics."""
        # euclidian, cosine, or city block from get_configs
        # calculate pdist
        dist_array = pdist(self.X, metric=self.get_config("metric", "euclidean"))
        dist_array = squareform(
            np.divide(  # normalize it st highest similarity is 1
                dist_array - np.amin(dist_array),
                np.amax(dist_array) - np.amin(dist_array),
            )
        )

        # at most, every row will be in its own cluster, so create a list of indices
        # and just iterate through it, skipping those that are already clustered
        rng = np.random.default_rng(self.get_config("random_state", None))
        idxs = np.arange(len(self.X))
        rng.shuffle(idxs)

        # use a default similarity or get_config
        distance_cutoff = self.get_config("distance_cutoff", 0.25)

        # build first cluster (add those that are clustered with it to a set as you go)
        # and continue until every row is sorted
        already_assigned = set()
        # output labels
        labels = np.full(len(self.X), fill_value=-1, dtype=int)
        # label counter
        cluster_idx = 0
        for sample_idx in idxs:
            if sample_idx in already_assigned:  # skip rows that are already assigned
                continue
            # get the row of the distance matrix for that sample
            row = dist_array[sample_idx]
            # find indices where it is close (given distance cutoff) including itself
            candidate_indices = set(np.flatnonzero(row < distance_cutoff))
            # check which of these have not already been assigned
            unassigned_indices = candidate_indices.difference(already_assigned)
            # add the labels to these
            for i in unassigned_indices:
                labels[i] = cluster_idx
            # increment the cluster index
            cluster_idx += 1
            # add these used indices to the tracker
            already_assigned.update(unassigned_indices)

        self._samples_clusters = labels



"""
This sampler partitions the data based on molecular weight. It first sorts the
molecules by molecular weight and then places the smallest molecules in the training set,
the next smallest in the validation set if applicable, and finally the largest molecules
in the testing set.
"""

import numpy as np

try:
    from astartes.utils.aimsim_featurizer import featurize_molecules
except ImportError:
    # this is in place so that the import of this from parent directory will work
    # if it fails, it is caught in molecules instead and the error is more helpful
    NO_MOLECULES = True

from .scaffold import Scaffold
from .target_property import TargetProperty


# inherit sample method from TargetProperty
class MolecularWeight(TargetProperty):
    def _before_sample(self):
        # check for invalid data types using the method in the Scaffold sampler
        Scaffold._validate_input(self.X)
        # calculate the average molecular weight of the molecule
        self.y_backup = self.y
        self.y = featurize_molecules((Scaffold.str_to_mol(i) for i in self.X), "mordred:MW", fprints_hopts={})

    def _after_sample(self):
        # restore the original y values
        self.y = self.y_backup



import numpy as np
from sklearn.model_selection import train_test_split as sk_train_test_split

from astartes.samplers import AbstractSampler


class Random(AbstractSampler):
    def _sample(self):
        """Passthrough to sklearn train_test_split"""
        idx_list = list(range(len(self.X)))
        train_indices, test_indices = sk_train_test_split(
            idx_list,
            train_size=len(idx_list) - 1,
            random_state=self.get_config("random_state", None),
            shuffle=self.get_config("shuffle", True),
        )
        self._samples_idxs = np.array(train_indices + test_indices, dtype=int)



import warnings

import numpy as np

from astartes.utils.exceptions import MoleculesNotInstalledError

Molecule = None
try:
    """
    aimsim depends on sklearn_extra, which uses a version checking technique that is due to
    be deprecated in a version of Python after 3.11, so it is throwing a deprecation warning
    We ignore this warning since we can't do anything about it (sklearn_extra seems to be
    abandonware) and in the future it will become an error that we can deal with.
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=DeprecationWarning)
        from aimsim.chemical_datastructures import Molecule
        from aimsim.exceptions import LoadingError
except ImportError:  # pragma: no cover
    pass  # Raise an error later


def featurize_molecules(molecules, fingerprint, fprints_hopts):
    """Call AIMSim's Molecule to featurize the molecules according to the arguments.

    Args:
        molecules (np.array): SMILES strings or RDKit molecule objects.
        fingerprint (str): The molecular fingerprint to be used.
        fprints_hopts (dict): Hyperparameters for AIMSim.

    Returns:
        np.array: X array (featurized molecules)
    """
    if Molecule is None:
        raise MoleculesNotInstalledError(
            """To use molecule featurizer, install astartes with pip install astartes[molecules] or conda install astartes aimsim."""
        )
    X = []
    for molecule in molecules:
        try:
            if type(molecule) in (np.str_, str):
                mol = Molecule(mol_smiles=molecule)
            else:
                mol = Molecule(mol_graph=molecule)
        except LoadingError as le:
            raise RuntimeError(
                "Unable to featurize molecules using '{:s}' with this configuration: fprint_hopts={:s}"
                "\nCheck terminal output for messages from the RDkit logger. ".format(fingerprint, repr(fprints_hopts))
            ) from le
        mol.descriptor.make_fingerprint(
            mol.mol_graph,
            fingerprint,
            fingerprint_params=fprints_hopts,
        )
        X.append(mol.descriptor.to_numpy())
    return np.array(X)



# import functions from this directory's contents so that users can import
# them with `from astartes.utils import *`
# internally, we do NOT do this to make the imports more explicit, i.e.
# `from astartes.utils.exceptions import *`
from .user_utils import generate_regression_results_dict

__all__ = ["generate_regression_results_dict"]



import numbers
import warnings
from datetime import date, datetime

import numpy as np
import pandas as pd

from astartes.utils.exceptions import InvalidConfigurationError, UncastableInputError
from astartes.utils.warnings import ConversionWarning


def convert_to_array(obj: object, name: str):
    """Attempt to convert obj named name to a numpy array, with appropriate warnings and exceptions.

    Args:
        obj (object): The item to attempt to convert.
        name (str): Human-readable name for printing.
    """
    warnings.warn(
        "Attempting to cast {:s} to a numpy array, which may result in unexpected behavior "
        "(remove this warning by passing numpy arrays directly to astartes).".format(
            name,
        ),
        ConversionWarning,
    )
    try:
        new_array = np.asarray(obj)
    except Exception as e:
        raise UncastableInputError(
            "Unable to cast {:s} to a numpy array using np.asarray."
        ) from e
    # ensure that all the values in the array are floats or date/datetime objects
    for item in new_array.ravel():
        if not hasattr(item, "__add__"):  # only allow items which we can do math on
            raise InvalidConfigurationError(
                "After casting, input object {:s} contained unsupported type {:s}".format(
                    name,
                    str(type(item)),
                )
            )

    return new_array


def panda_handla(X, y, labels):
    """Helper function to deal with supporting Pandas data types in astartes

    Args:
        X (Dataframe): Features with column names
        y (Series): Targets
        labels (Series): Labels for data

    Returns:
        dict: Empty if no pandas types, metadata-filled otherwise
    """
    output_is_pandas = {}
    # X can be a dataframe, in which case save the column names
    if type(X) is pd.DataFrame:
        output_is_pandas["X"] = {"columns": X.columns, "index": X.index}

    # y can be a series, indicate with True
    if type(y) is pd.Series:  # this implicitly avoids None types
        output_is_pandas["y"] = {"index": y.index}

    # labels can also be a series, indicate with True
    if type(labels) is pd.Series:
        output_is_pandas["labels"] = {"index": labels.index}

    return output_is_pandas


def return_helper(
    sampler_instance,
    train_idxs,
    val_idxs,
    test_idxs,
    return_indices,
    output_is_pandas,
):
    """Convenience function to return the requested arrays appropriately.

    Args:
        sampler_instance (sampler): The fit sampler instance.
        test_size (float): Fraction of data to use in test.
        val_size (float): Fraction of data to use in val.
        train_size (float): Fraction of data to use in train.
        return_indices (bool): Return indices after the value arrays.
        output_is_pandas (dict): metadata about casting to pandas.

    Returns:
        np.array: Either many arrays or indices in arrays.

    Notes:
        This function copies and pastes a lot of code when it could instead
        use some loop over (X, y, labels, sampler_instance.get_clusters())
        but such an implementation is more error prone. This is long and
        not the prettiest, but it is definitely doing what we want.
    """
    # pack all the output into a list and then unpack it in the return statement
    out = []

    # Feature array
    X_train = sampler_instance.X[train_idxs]
    # check if it needs to be converted back to a Dataframe
    dataframe_metadata = output_is_pandas.get("X", False)
    if dataframe_metadata:
        X_train = pd.DataFrame(
            X_train,
            columns=dataframe_metadata["columns"],
            index=dataframe_metadata["index"][train_idxs],
        )
    out.append(X_train)
    # repeat the above process for validation and test
    if len(val_idxs):
        X_val = sampler_instance.X[val_idxs]
        if dataframe_metadata:
            X_val = pd.DataFrame(
                X_val,
                columns=dataframe_metadata["columns"],
                index=dataframe_metadata["index"][val_idxs],
            )
        out.append(X_val)
    X_test = sampler_instance.X[test_idxs]
    if dataframe_metadata:
        X_test = pd.DataFrame(
            X_test,
            columns=dataframe_metadata["columns"],
            index=dataframe_metadata["index"][test_idxs],
        )
    out.append(X_test)

    # repeat above with y, if it exists
    if sampler_instance.y is not None:
        y_train = sampler_instance.y[train_idxs]
        # if user passed y as a series, recast it
        series_metadata = output_is_pandas.get("y", False)
        if series_metadata:
            y_train = pd.Series(
                y_train,
                index=series_metadata["index"][train_idxs],
            )
        out.append(y_train)
        if len(val_idxs):
            y_val = sampler_instance.y[val_idxs]
            if series_metadata:
                y_val = pd.Series(
                    y_val,
                    index=series_metadata["index"][val_idxs],
                )
            out.append(y_val)
        y_test = sampler_instance.y[test_idxs]
        if series_metadata:
            y_test = pd.Series(
                y_test,
                index=series_metadata["index"][test_idxs],
            )
        out.append(y_test)

    # repeat with labels, if they exist
    if sampler_instance.labels is not None:
        labels_train = sampler_instance.labels[train_idxs]
        # if user passed labels as a series, recast it
        series_metadata = output_is_pandas.get("labels", False)
        if series_metadata:
            labels_train = pd.Series(
                labels_train,
                index=series_metadata["index"][train_idxs],
            )
        out.append(labels_train)
        if len(val_idxs):
            labels_val = sampler_instance.labels[val_idxs]
            if series_metadata:
                labels_val = pd.Series(
                    labels_val,
                    index=series_metadata["index"][val_idxs],
                )
            out.append(labels_val)
        labels_test = sampler_instance.labels[test_idxs]
        if series_metadata:
            labels_test = pd.Series(
                labels_test,
                index=series_metadata["index"][test_idxs],
            )
        out.append(labels_test)

    # repeat above with the clusters, if a clustering approach was used
    if len(sampler_instance.get_clusters()):  # true when the list has been filled
        clusters_train = sampler_instance.get_clusters()[train_idxs]
        out.append(clusters_train)
        if len(val_idxs):
            clusters_val = sampler_instance.get_clusters()[val_idxs]
            out.append(clusters_val)
        clusters_test = sampler_instance.get_clusters()[test_idxs]
        out.append(clusters_test)

    # append indices, if requested
    if return_indices:
        out.append(train_idxs)
        if val_idxs.any():
            out.append(val_idxs)
        out.append(test_idxs)

    return (*out,)



# abstract base classes
from .abstract_sampler import AbstractSampler

# implementations
from .extrapolation import (
    DBSCAN,
    KMeans,
    MolecularWeight,
    OptiSim,
    Scaffold,
    SphereExclusion,
    TargetProperty,
    TimeBased,
)
from .interpolation import SPXY, KennardStone, Random

IMPLEMENTED_INTERPOLATION_SAMPLERS = (
    "random",
    "kennard_stone",
    "spxy",
)

IMPLEMENTED_EXTRAPOLATION_SAMPLERS = (
    "dbscan",
    "scaffold",
    "kmeans",
    "molecular_weight",
    "optisim",
    "sphere_exclusion",
    "time_based",
    "target_property",
)

ALL_SAMPLERS = IMPLEMENTED_EXTRAPOLATION_SAMPLERS + IMPLEMENTED_INTERPOLATION_SAMPLERS

DETERMINISTIC_EXTRAPOLATION_SAMPLERS = (
    "time_based",
    "target_property",
    "molecular_weight",
)



from sklearn.cluster import DBSCAN as sk_DBSCAN

from astartes.samplers import AbstractSampler


class DBSCAN(AbstractSampler):
    def _sample(self):
        """
        Implements the DBSCAN sampler to identify clusters.
        Density-Based Spatial Clustering of Applications with Noise finds
        core samples in regions of high density and expands clusters from them.
        This algorithm is good for data which contains clusters of similar density.
        """

        dbscan = sk_DBSCAN(
            eps=self.get_config("eps", 0.5),
            min_samples=self.get_config("min_samples", 2),
            metric=self.get_config("metric", "euclidean"),
            algorithm=self.get_config("algorithm", "auto"),
            leaf_size=self.get_config("leaf_size", 30),
            p=self.get_config("p", None),
        ).fit(self.X)
        self._samples_clusters = dbscan.labels_



from math import floor
from typing import Union
from warnings import warn

import numpy as np
import pandas as pd

from astartes.samplers import (
    DETERMINISTIC_EXTRAPOLATION_SAMPLERS,
    IMPLEMENTED_EXTRAPOLATION_SAMPLERS,
    IMPLEMENTED_INTERPOLATION_SAMPLERS,
)
from astartes.utils.array_type_helpers import (
    convert_to_array,
    panda_handla,
    return_helper,
)
from astartes.utils.exceptions import InvalidConfigurationError
from astartes.utils.sampler_factory import SamplerFactory
from astartes.utils.warnings import ImperfectSplittingWarning, NormalizationWarning

# define random seed
DEFAULT_RANDOM_STATE = 42


def train_val_test_split(
    X: Union[np.array, pd.DataFrame],
    y: Union[np.array, pd.Series] = None,
    labels: Union[np.array, pd.Series] = None,
    train_size: float = 0.8,
    val_size: float = 0.1,
    test_size: float = 0.1,
    sampler: str = "random",
    random_state: int = None,
    hopts: dict = {},
    return_indices: bool = False,
):
    """Deterministic train_test_splitting of arbitrary arrays.

    Args:
        X (np.array, pd.DataFrame): Numpy array or pandas DataFrame of feature vectors.
        y (np.array, pd.Series, optional): Targets corresponding to X, must be of same size. Defaults to None.
        labels (np.array, pd.Series, optional): Labels corresponding to X, must be of same size. Defaults to None.
        train_size (float, optional): Fraction of dataset to use in training set. Defaults to 0.8.
        val_size (float, optional): Fraction of dataset to use in validation set. Defaults to 0.1.
        test_size (float, optional): Fraction of dataset to use in test set. Defaults to 0.1.
        sampler (str, optional): Sampler to use, see IMPLEMENTED_INTER/EXTRAPOLATION_SAMPLERS. Defaults to "random".
        random_state (int, optional): The random seed used throughout astartes.
        hopts (dict, optional): Hyperparameters for the sampler used above. Defaults to {}.
        return_indices (bool, optional): True to return indices of train/test after values. Defaults to False.

    Returns:
        np.array(s): X, y, and labels train/val/test data, or indices.
    """
    # special case for casting back to pandas
    output_is_pandas = panda_handla(X, y, labels)

    # now convert everything to numpy arrays for our internal algorithms
    if type(X) is not np.ndarray:
        X = convert_to_array(X, "X")
    if y is not None and type(y) is not np.ndarray:
        y = convert_to_array(y, "y")
    if labels is not None and type(labels) is not np.ndarray:
        labels = convert_to_array(labels, "labels")

    # check for consistent length after conversion
    msg = ""
    if y is not None and len(y) != len(X):
        msg += "len(y)={:d} ".format(len(y))
    if labels is not None and len(labels) != len(X):
        msg += "len(labels)={:d} ".format(len(labels))
    if msg:
        raise InvalidConfigurationError("Lengths of input arrays do not match: len(X)={:d} ".format(len(X)) + msg)

    train_size, val_size, test_size = _normalize_split_sizes(train_size, val_size, test_size)
    hopts["random_state"] = random_state if random_state is not None else DEFAULT_RANDOM_STATE
    sampler_factory = SamplerFactory(sampler)
    sampler_instance = sampler_factory.get_sampler(X, y, labels, hopts)

    if sampler in (
        *IMPLEMENTED_INTERPOLATION_SAMPLERS,
        *DETERMINISTIC_EXTRAPOLATION_SAMPLERS,
    ):
        # extrapolation samplers in DETERMINISTIC_EXTRAPOLATION_SAMPLERS do not accept the
        # random_state argument because they are entirely deterministic
        return _interpolative_sampling(
            sampler_instance,
            test_size,
            val_size,
            train_size,
            return_indices,
            output_is_pandas,
        )
    else:
        return _extrapolative_sampling(
            sampler_instance,
            test_size,
            val_size,
            train_size,
            return_indices,
            output_is_pandas,
            random_state,
        )


def train_test_split(
    X: np.array,
    y: np.array = None,
    labels: np.array = None,
    train_size: float = 0.75,
    test_size: float = None,
    sampler: str = "random",
    random_state: int = None,
    hopts: dict = {},
    return_indices: bool = False,
):
    """Deterministic train_test_splitting of arbitrary arrays.

    Args:
        X (np.array): Numpy array of feature vectors.
        y (np.array, optional): Targets corresponding to X, must be of same size. Defaults to None.
        labels (np.array, optional): Labels corresponding to X, must be of same size. Defaults to None.
        train_size (float, optional): Fraction of dataset to use in training set. Defaults to 0.75.
        test_size (float, optional): Fraction of dataset to use in test set. Defaults to None.
        sampler (str, optional): Sampler to use, see IMPLEMENTED_INTER/EXTRAPOLATION_SAMPLERS. Defaults to "random".
        random_state (int, optional): The random seed used throughout astartes.
        hopts (dict, optional): Hyperparameters for the sampler used above. Defaults to {}.
        return_indices (bool, optional): True to return indices of train/test instead of values. Defaults to False.

    Returns:
        np.array: X, y, and labels train/val/test data, or indices.
    """
    return train_val_test_split(
        X,
        y,
        labels,
        train_size,
        0,
        test_size,
        sampler,
        random_state,
        hopts,
        return_indices,
    )


def _extrapolative_sampling(
    sampler_instance,
    test_size,
    val_size,
    train_size,
    return_indices,
    output_is_pandas,
    random_state,
):
    """Helper function to perform extrapolative sampling.

    Attempts to fill train, val, and test without breaking up clusters. Prioiritizes underfilling
    test and then val and then the balance goes into train (which is why the floats are given
    in that order).

    Args:
        sampler_instance (sampler): The fit sampler instance.
        test_size (float): Fraction of data to use in test.
        val_size (float): Fraction of data to use in val.
        train_size (float): Fraction of data to use in train.
        return_indices (bool): Return indices or the arrays themselves.
        output_is_pandas (array[str] or bool): True/False if output should cast to pandas,
            data needed to perform casting if True.
        random_state (int, optional): The random state used to shuffle small clusters. Default to no shuffle.

    Returns:
        calls: return_helper
    """
    # calculate "goal" splitting sizes
    n_train_samples = floor(len(sampler_instance.X) * train_size)
    n_val_samples = floor(len(sampler_instance.X) * val_size)
    n_test_samples = floor(len(sampler_instance.X) * test_size)

    if val_size == 0:
        max_shufflable_size = min(n_train_samples, n_test_samples)
    else:
        # typically, the test set and val set are smaller than the training set
        max_shufflable_size = min(n_test_samples, n_val_samples)
    # unlike interpolative, cannot calculate n_train_samples here
    # since it will vary based on cluster_lengths

    # largest clusters must go into largest set, but smaller ones can optionally
    # be shuffled
    cluster_counter = sampler_instance.get_sorted_cluster_counter(max_shufflable_size=max_shufflable_size if random_state is not None else None)

    test_idxs, val_idxs, train_idxs = (
        np.array([], dtype=int),
        np.array([], dtype=int),
        np.array([], dtype=int),
    )
    for cluster_idx, cluster_length in cluster_counter.items():
        # assemble test first, avoid overfilling
        if (len(test_idxs) + cluster_length) <= n_test_samples:
            test_idxs = np.append(test_idxs, sampler_instance.get_sample_idxs(cluster_length))
        elif (len(val_idxs) + cluster_length) <= n_val_samples:
            val_idxs = np.append(val_idxs, sampler_instance.get_sample_idxs(cluster_length))
        else:  # then balance goes into train
            train_idxs = np.append(train_idxs, sampler_instance.get_sample_idxs(cluster_length))
    _check_actual_split(train_idxs, val_idxs, test_idxs, train_size, val_size, test_size)
    return return_helper(
        sampler_instance,
        train_idxs,
        val_idxs,
        test_idxs,
        return_indices,
        output_is_pandas,
    )


def _interpolative_sampling(
    sampler_instance,
    test_size,
    val_size,
    train_size,
    return_indices,
    output_is_pandas,
):
    """Helper function to perform interpolative sampling.

    Attempts to fill train, val, and test within rounding limits. Prioiritizes underfilling
    train and then val and then the balance goes into test (this is the opposite to extrapolative
    because these samplers move outisde-in).

    Args:
        sampler_instance (sampler): The fit sampler instance.
        test_size (float): Fraction of data to use in test.
        val_size (float): Fraction of data to use in val.
        train_size (float): Fraction of data to use in train.
        return_indices (bool): Return indices or the arrays themselves.
        output_is_pandas (array[str] or bool): True/False if output should cast to pandas,
            data needed to perform casting if True.

    Returns:
        calls: return_helper
    """
    n_train_samples = floor(len(sampler_instance.X) * train_size)
    n_val_samples = floor(len(sampler_instance.X) * val_size)
    n_test_samples = len(sampler_instance.X) - (n_train_samples + n_val_samples)

    train_idxs = sampler_instance.get_sample_idxs(n_train_samples)
    val_idxs = sampler_instance.get_sample_idxs(n_val_samples)
    test_idxs = sampler_instance.get_sample_idxs(n_test_samples)

    _check_actual_split(train_idxs, val_idxs, test_idxs, train_size, val_size, test_size)
    return return_helper(
        sampler_instance,
        train_idxs,
        val_idxs,
        test_idxs,
        return_indices,
        output_is_pandas,
    )


def _check_actual_split(
    train_idxs,
    val_idxs,
    test_idxs,
    train_size,
    val_size,
    test_size,
):
    """Check for empty sets and imperfect splits in the results.

    Args:
        train_idxs (np.array): The resulting training indices from sampling.
        val_idxs (np.array): Validation indices.
        test_idxs (np.array): Testing indices.
        train_size (float): The user-requested (or normalized) training fraction.
        val_size (float): The user-requested (or normalized) validation fraction.
        test_size (float): The user-requested (or normalized) testing fraction.

    Raises:
        InvalidConfigurationError: Raised when a set turns out empty due to configuration.
    """
    # split may have resulted in empty lists
    msg = ""
    if not len(train_idxs):
        msg += " training set empty "
    if not len(test_idxs):
        msg += " testing set empty "
    if val_size and not len(val_idxs):
        msg += " validation set empty "
    if msg:
        raise InvalidConfigurationError(
            "Provided data and requested split resulted in an empty set. "
            "Dataset may be too small or requested splits may be too large. "
            "(" + msg + ")"
        )
    # split may not match exactly what they asked for
    total_size = len(test_idxs) + len(val_idxs) + len(train_idxs)

    actual_train_size = round(len(train_idxs) / total_size, 2)
    requested_train_size = round(train_size, 2)

    actual_val_size = round(len(val_idxs) / total_size, 2)
    requested_val_size = round(val_size, 2)

    actual_test_size = round(len(test_idxs) / total_size, 2)
    requested_test_size = round(test_size, 2)

    msg = ""
    if actual_train_size != requested_train_size:
        msg += "Requested train size of {:.2f}, got {:.2f}. ".format(
            requested_train_size,
            actual_train_size,
        )
    if actual_val_size != requested_val_size:
        msg += "Requested validation size of {:.2f}, got {:.2f}. ".format(
            requested_test_size,
            actual_test_size,
        )
    if actual_test_size != requested_test_size:
        msg += "Requested test size of {:.2f}, got {:.2f}. ".format(
            requested_test_size,
            actual_test_size,
        )
    if msg:
        warn(
            "Actual train/test split differs from requested size. " + msg,
            ImperfectSplittingWarning,
        )


def _normalize_split_sizes(train_size, val_size, test_size):
    """Normalize requested inputs to between zero and one (summed) with extensive validation."""
    if not train_size and not test_size:  # neither - error
        raise InvalidConfigurationError(
            "train_size or test_size must be nonzero (val_size will default to 0.0).\n"
            "(got val_size={:s} test_size={:s} train_size={:s})".format(
                repr(val_size),
                repr(test_size),
                repr(train_size),
            )
        )
    out = []
    if not val_size:  # doing train_test_split
        if train_size and test_size:  # both - normalize
            if train_size + test_size != 1.0:
                out_train_size = train_size / (train_size + test_size)
                out_test_size = 1.0 - out_train_size
                warn(
                    "Requested train/test split ({:.2f}, {:.2f}) do not sum to 1.0,"
                    " normalizing to train={:.2f}, test={:.2f}.".format(
                        train_size,
                        test_size,
                        out_train_size,
                        out_test_size,
                    ),
                    NormalizationWarning,
                )
                out = [out_train_size, 0.0, out_test_size]
            else:
                out = [train_size, val_size, test_size]
        else:  # one or the other - only allow floats [0, 1), then calculate
            if train_size:
                if train_size >= 1.0 or train_size <= 0:
                    raise InvalidConfigurationError("If specifying only train_size, must be float between (0, 1) (got {:.2f})".format(train_size))
                test_size = 1.0 - train_size
                out = [train_size, 0, test_size]
            else:
                if test_size >= 1.0 or test_size <= 0:
                    raise InvalidConfigurationError("If specifying only test_size, must be float between (0, 1) (got {:.2f})".format(test_size))
                train_size = 1.0 - test_size
                out = [train_size, 0, test_size]
    else:  # there is a non-zero val_size
        if val_size >= 1.0 or val_size <= 0.0:
            raise InvalidConfigurationError("val_size must be a float between (0, 1) (got {:.2f})".format(val_size))
        if train_size and test_size:  # all three - normalize
            if train_size + test_size + val_size != 1.0:
                normalization = train_size + test_size + val_size
                out_train_size = train_size / normalization
                out_test_size = test_size / normalization
                out_val_size = val_size / normalization
                warn(
                    "Requested train/val/test split ({:.2f}, {:.2f}, {:.2f}) do not sum to 1.0,"
                    " normalizing to train={:.2f}, val={:.2f}, test={:.2f}.".format(
                        train_size,
                        val_size,
                        test_size,
                        out_train_size,
                        out_val_size,
                        out_test_size,
                    ),
                    NormalizationWarning,
                )
                out = [out_train_size, val_size, out_test_size]
            else:
                out = [train_size, val_size, test_size]
        else:  # one or the other - only allow floats [0, 1), then calculate
            if train_size:
                if train_size >= 1.0 or train_size <= 0:
                    raise InvalidConfigurationError(
                        "If specifying val_size and only train_size, must be float between (0, 1) (got {:.2f})".format(train_size)
                    )
                test_size = 1.0 - (train_size + val_size)
                out = [train_size, val_size, test_size]
            else:
                if test_size >= 1.0 or test_size <= 0:
                    raise InvalidConfigurationError(
                        "If specifying val_size and only test_size, must be float between (0, 1) (got {:.2f})".format(test_size)
                    )
                train_size = 1.0 - (test_size + val_size)
                out = [train_size, val_size, test_size]

    return (*out,)



"""Warnings used by astartes"""


class ImperfectSplittingWarning(RuntimeWarning):
    """Used when a sampler cannot match requested splits."""

    def __init__(self, message=None):
        self.message = message
        super().__init__(message)


class NormalizationWarning(RuntimeWarning):
    """Used when a requested split does not add to 1."""

    def __init__(self, message=None):
        self.message = message
        super().__init__(message)


class ConversionWarning(RuntimeWarning):
    """Used when passed data is not a numpy array."""

    def __init__(self, message=None):
        self.message = message
        super().__init__(message)


class NoMatchingScaffold(Warning):
    """
    Used when an RDKit molecule does not match any
    Bemis-Murcko scaffold and returns an empty string.
    """

    def __init__(self, message=None):
        self.message = message
        super().__init__(message)



"""
The Optimizable K-Dissimilarity Selection (OptiSim) algorithm, as originally
described by Clark (https://pubs.acs.org/doi/full/10.1021/ci970282v), adapted
to work for arbitrary distance metrics.

The original algorithm:
1. Initialization
 - Take a featurized dataset and select an arbitrary starting data point for
   the selection set.
 - Treat the remaining data as 'candidates'.
 - Create an empty 'recycling bin'.
 - Create an empty subsample set.
 - Create an empty selection set.

2. Remove a random point from the candidates.
 - if it has a similarity greater than a given cutoff to any of the members of the selection set,
 recycle it (or conversely, if it is within a cutoff distance)
 - otherwise, add to subsample set

3. Repeat 2 until one of two conditions is met:
 a. The subsample reaches the pre-determined maximum size K or
 b. The candidates are exhausted.

4. If Step 3 resulted in condition b, move all data from recycling bin and go
to Step 2.

5. If subsample is empty, quit (all remaining candidates are similar, the
most dissimilar data points have already been identified)

6. Pick the most dissimilar (relative to data points already in selection set)
point in the subsample and add it to the selection set.

7. Move the remaining points in the subsample to the recycling bin.

8. If size(selection set) is sufficient, quit. Otherwise, go to Step 2.

As suggested in the original paper, the members of the selection set are then
used as cluster centers, and we assign every element in the dataset to belong
to the cluster containing the selection set member to which it is the most
similar. To implement this step, use scipy.spatial.distance.cdist.

This algorithm seems like it might introduce an infinite loop if the subsample
is not filled and all of the remaining candidates are within the cutoff and cannot
be added. Might need a stop condition here? Unless the empyting of the recycling bin
will somehow fix this. Also possible that one would never have a partially filled
subsample after looking at the full dataset since it is more probable that ALL the
points would be rejected and the subsample would be empty.

Likely just check for no more points being possible to fit into the subsample, and
exit if that is the case.

"""
from math import floor

import numpy as np
from scipy.spatial.distance import pdist

from astartes.samplers import AbstractSampler


class OptiSim(AbstractSampler):
    def _sample(self):
        """Implementes the OptiSim sampler"""
        self._init_random(self.get_config("random_state", 42))
        n_samples = len(self.X)

        # pull out the user configs, named as they are in the original algorithm
        # the max number of clusters to find
        M = self.get_config("n_clusters", 10)
        # use up to 5% of the data per iteration in sampler
        K = self.get_config("max_subsample_size", 1 + floor(n_samples * 0.05))
        # up to 10% distance is good enough for default
        c = self.get_config("distance_cutoff", 0.10)

        # all of the sets described in the paper will operate with indices
        selection_set = set()
        subsample_set = set()
        recycling_bin = set()
        candidate_set = set(range(n_samples))

        # pick an arbitrary first member of the selection set
        initial_selection = self.rchoose(candidate_set)
        self.move_item(initial_selection, candidate_set, selection_set)

        # pick an arbitrary starting point, add to subsample, remove from candidates
        start_sample = self.rchoose(candidate_set)
        self.move_item(start_sample, candidate_set, subsample_set)

        # Step 5 in the process is rare and does not fit well with the below implementation
        # since it would require nested while loops. Thus, we will introduce an break counter
        # and use it as a backup (proboably good policy anyway).
        emergency_break = 0
        # worst case, we try putting each sample into each cluster once
        _it_limit = M * n_samples

        # continue as long as selection set is not full (and avoid infinite loop)
        while len(selection_set) < M and emergency_break < _it_limit:
            if len(subsample_set) == K:  # 3a
                # pick subsample member most dissimilar to the selection set
                # by adding the distance of each subsample member from
                # the members of the selection set
                distance_scores = {}
                for sample in subsample_set:
                    distance_scores[sample] = 0.0
                    for selection_sample in selection_set:
                        distance_scores[sample] += self.get_dist(
                            sample, selection_sample
                        )
                furthest_sample = max(distance_scores, key=distance_scores.get)
                self.move_item(furthest_sample, subsample_set, selection_set)
                # move other subsampled items to recycling, clear subsample
                recycling_bin.update(subsample_set)
                subsample_set.clear()
            elif not len(candidate_set):  # 3b
                # move data from recycling to candidate list
                candidate_set.update(recycling_bin)
                recycling_bin.clear()
            else:  # subsample not full, still more candidates, attempt to add
                candidate_sample = self.rchoose(candidate_set)
                if any(  # it is different from all the
                    [self.get_dist(candidate_sample, i) < c for i in selection_set]
                ):
                    self.move_item(candidate_sample, candidate_set, recycling_bin)
                else:
                    self.move_item(candidate_sample, candidate_set, subsample_set)

            emergency_break += 1

        # now we assign the clusters based on which samples are closest to which members
        # of the selection set
        cluster_labels = dict(zip(selection_set, range(len(selection_set))))
        labels = np.full(n_samples, fill_value=-1, dtype=int)
        for sample in range(n_samples):
            if sample in cluster_labels.keys():  # cluster center
                labels[sample] = cluster_labels[sample]
            else:  # find which cluster center is closest
                distances = {
                    i: self.get_dist(sample, i) for i in cluster_labels.values()
                }
                labels[sample] = min(distances, key=distances.get)

        self._samples_clusters = labels

    def _init_random(self, random_state):
        """This uses a lot of random numbers, so make them convenient to reach."""
        self._rng = np.random.default_rng(seed=random_state)
        return

    def rchoose(self, set):
        """Choose a random element from a set with self._rng"""
        return self._rng.choice(list(set))

    def get_dist(self, i, j):
        """Calculates pdist and returns distance between two samples"""
        if not hasattr(self, "_pdist"):
            # calculate the distance matrix, normalize it
            dist_array = pdist(self.X, metric=self.get_config("metric", "euclidean"))
            self._pdist = np.divide(
                dist_array - np.amin(dist_array),
                np.amax(dist_array) - np.amin(dist_array),
            )
        return self._pdist[len(self.X) * i + j - ((i + 2) * (i + 1)) // 2]

    def move_item(self, item, source_set, destintation_set):
        """Moves item from source_set to destination_set"""
        destintation_set.add(item)
        source_set.remove(item)
        return



from .dbscan import DBSCAN
from .kmeans import KMeans
from .molecular_weight import MolecularWeight
from .optisim import OptiSim
from .scaffold import Scaffold
from .sphere_exclusion import SphereExclusion
from .target_property import TargetProperty
from .time_based import TimeBased



"""Abstract Sampling class"""

from abc import ABC, abstractmethod
from collections import Counter

import numpy as np


class AbstractSampler(ABC):
    """
    Abstract Base Class for all samplers.
    """

    def __init__(self, X, y, labels, configs):
        """Copies X, y, labels, and configs into class attributes and then calls sampler."""
        # required in all splitting methods
        self.X = X

        # optional in all splitting methods
        self.y = y
        self.labels = labels
        self._configs = configs

        # this must be set by _sample for interpolation
        self._samples_idxs = np.array([], dtype=int)

        # this must be set if using a clustering algorithm
        self._samples_clusters = np.array([], dtype=int)

        # internal machinery
        self._current_sample_idx = 0
        self._before_sample()
        self._sample()
        self._after_sample()

    def _before_sample(self):
        """This method should perform any data validation, manipulation, etc. required before proceeding to _sample

        Returns:
            None: Returns nothing, raises an Exception if something is wrong.
        """

    def _after_sample(self):
        """This method should perform any checks, mutations, etc. required after _sample is completed.

        Returns:
            None: Returns nothing, raises an Exception if something is wrong.
        """

    @abstractmethod
    def _sample(self):
        """
        This method should: (arrays should be np.ndarray)
         - set self._samples_idxs with the order in which the algorithm dictates drawing samples
        and if using clustering:
         - set self._samples_clusters with the labels produced by clustering
        """

    def get_config(self, key, default=None):
        """Getter to sampler._configs

        Args:
            key (str): String parameter for the sampler.
            default (any, optional): Default to return if key not present. Defaults to None.

        Returns:
            any: value at provided key, or else default.
        """
        return self._configs.get(key, default)

    def get_sample_idxs(self, n_samples):
        """
        Get idxs of samples.
        """
        out = self._samples_idxs[self._current_sample_idx : self._current_sample_idx + n_samples]
        self._current_sample_idx += n_samples
        return out

    def get_clusters(self):
        """Getter for the cluster labels.

        Returns:
            np.array: Cluster labels.
        """
        return self._samples_clusters

    def get_sorted_cluster_counter(self, max_shufflable_size=None):
        """
        Return a dict containing cluster_id: number of members sorted by number
        of members, ascending

        if max_shufflable_size is not None, clusters below the passed size will be
        shuffled into a new order according to random_state in hopts
        """
        # dictionary of cluster label: length (number of members)
        cluster_counter = Counter(self._samples_clusters)

        # same, sorted by number of members ascending
        ordered_cluster_counter = dict(sorted(cluster_counter.items(), key=lambda i: i[1]))

        # if using max_shufflable_size, shuffle the small clusters in this dictionary
        if max_shufflable_size is not None:
            # get a list of the small clusters
            small_clusters = [cluster_label for cluster_label, length in ordered_cluster_counter.items() if length <= max_shufflable_size]

            # the remaining clusters go here
            large_clusters = [i for i in ordered_cluster_counter.keys() if i not in small_clusters]

            # shuffle the small clusters according to the random state
            rng = np.random.default_rng(seed=self.get_config("random_state"))
            rng.shuffle(small_clusters)

            # recombine the clusters, rebuild the ordered dictionary
            all_clusters = small_clusters + large_clusters
            ordered_cluster_counter = {label: cluster_counter[label] for label in all_clusters}

        # make a new dictionary that maps cluster label: indexes of members
        ordered_cluster_counter_with_idxs = {
            cluster_label: np.where(self._samples_clusters == cluster_label)[0] for cluster_label, length in ordered_cluster_counter.items()
        }

        # put all the indices in order
        self._samples_idxs = np.hstack(tuple(ordered_cluster_counter_with_idxs.values()))

        return ordered_cluster_counter



"""
This sampler partitions the data based on the Bemis-Murcko scaffold function as implemented in RDKit.
Bemis, G. W.; Murcko, M. A. The Properties of Known Drugs. 1. Molecular Frameworks. J. Med. Chem. 1996, 39, 2887−2893.
Landrum, G. et al. RDKit: Open-Source Cheminformatics; 2006; https://www.rdkit.org.

The goal is to cluster molecules that share the same scaffold.
Later, these clusters will be assigned to training, validation, and testing split
to create data splits that will measure extrapolation by testing on scaffolds
that are not in the training set.

"""

import warnings
from collections import defaultdict

import numpy as np

try:
    from rdkit import Chem
    from rdkit.Chem.Scaffolds import MurckoScaffold
except ImportError:
    # this is in place so that the import of this from parent directory will work
    # if it fails, it is caught in molecules instead and the error is more helpful
    NO_MOLECULES = True

from astartes.samplers import AbstractSampler
from astartes.utils.warnings import NoMatchingScaffold


class Scaffold(AbstractSampler):
    def _validate_input(X):
        # ensure that X contains entries that are either a SMILES string or an RDKit Molecule
        if not all(isinstance(i, str) for i in X) and not all(isinstance(i, Chem.rdchem.Mol) for i in X):
            msg = "Scaffold class requires input X to be an iterable of SMILES strings, InChI strings, or RDKit Molecules"
            raise TypeError(msg)

    def _before_sample(self):
        Scaffold._validate_input(self.X)

    def _sample(self):
        """Implements the Scaffold sampler to identify clusters via a molecule's Bemis-Murcko scaffold."""
        scaffold_to_indices = self.scaffold_to_smiles(self.X)

        cluster_indices = np.empty(len(self.X), dtype=object)
        # give each cluster an arbitrary ID
        for cluster_id, (scaffold, indices) in enumerate(scaffold_to_indices.items()):
            if scaffold == "":
                warnings.warn(
                    f"No matching scaffold was found for the {len(indices)} " f"molecules corresponding to indices {indices}",
                    NoMatchingScaffold,
                )
            for idx in indices:
                cluster_indices[idx] = scaffold

        self._samples_clusters = cluster_indices

    def scaffold_to_smiles(self, mols):
        """
        Computes scaffold for each smiles string and returns a mapping from scaffolds to sets of smiles.

        Params:
            mols: A list of smiles strings or RDKit molecules.

        Returns:
            A dictionary mapping each unique scaffold to all smiles (or smiles indices) which have that scaffold.
        """
        scaffolds = defaultdict(set)
        for i, mol in enumerate(mols):
            scaffold = Scaffold.generate_bemis_murcko_scaffold(mol, self.get_config("include_chirality", False))
            scaffolds[scaffold].add(i)

        return scaffolds

    def str_to_mol(string):
        """
        Converts an InChI or SMILES string to an RDKit molecule.

        Params:
            string: The InChI or SMILES string.

        Returns:
            An RDKit molecule.
        """
        RDKIT_SMILES_PARSER_PARAMS = Chem.SmilesParserParams()
        if string.startswith("InChI"):
            mol = Chem.MolFromInchi(string, removeHs=True)
        else:
            # Set params here so we don't remove hydrogens with atom mapping
            RDKIT_SMILES_PARSER_PARAMS.removeHs = True
            mol = Chem.MolFromSmiles(string, RDKIT_SMILES_PARSER_PARAMS)

        # atom map numbers should not be present when creating scaffolds
        for atom in mol.GetAtoms():
            atom.SetAtomMapNum(0)

        return mol

    def generate_bemis_murcko_scaffold(mol, include_chirality=False):
        """
        Compute the Bemis-Murcko scaffold for an RDKit molecule.

        Params:
            mol: A smiles string or an RDKit molecule.
            include_chirality: Whether to include chirality.

        Returns:
            Bemis-Murcko scaffold
        """
        mol = Scaffold.str_to_mol(mol) if isinstance(mol, str) else mol
        scaffold = MurckoScaffold.MurckoScaffoldSmiles(mol=mol, includeChirality=include_chirality)

        return scaffold



from scipy.spatial.distance import pdist, squareform

from astartes.samplers import AbstractSampler
from astartes.utils.fast_kennard_stone import fast_kennard_stone


class KennardStone(AbstractSampler):
    def _sample(self):
        """
        Implements the Kennard-Stone algorithm
        """
        self._samples_idxs = fast_kennard_stone(
            squareform(
                pdist(
                    self.X,
                    self.get_config("metric", "euclidean"),
                )
            )
        )



"""Exceptions used by astartes"""


class MoleculesNotInstalledError(RuntimeError):  # pragma: no cover
    """Used when attempting to featurize molecules without install."""

    def __init__(self, message=None):
        self.message = message
        super().__init__(message)


class InvalidModelTypeError(RuntimeError):
    """Used when user-provided model is invalid."""

    def __init__(self, message=None):
        self.message = message
        super().__init__(message)


class InvalidConfigurationError(RuntimeError):
    """Used when user-requested split/data would not work."""

    def __init__(self, message=None):
        self.message = message
        super().__init__(message)


class SamplerNotImplementedError(RuntimeError):
    """Used when attempting to call a non-existent sampler."""

    def __init__(self, message=None):
        self.message = message
        super().__init__(message)


class UncastableInputError(RuntimeError):
    """Used when X, y, or labels cannot be cast to a np.array."""

    def __init__(self, message=None):
        self.message = message
        super().__init__(message)



import numpy as np


def fast_kennard_stone(ks_distance: np.ndarray) -> np.ndarray:
    """Implements the Kennard-Stone algorithm

    Args:
        ks_distance (np.ndarray): Distance matrix

    Returns:
        np.ndarray: Indices in order of Kennard-Stone selection
    """
    n_samples = len(ks_distance)

    # when searching for max distance, disregard self
    np.fill_diagonal(ks_distance, -np.inf)

    # get the row/col of maximum (greatest distance)
    max_idx = np.nanargmax(ks_distance)
    max_coords = np.unravel_index(max_idx, ks_distance.shape)

    # list of indices which have been selected
    # - used to mask ks_distance
    # - also tracks order of kennard-stone selection
    already_selected = np.empty(n_samples, dtype=int)
    already_selected[0] = max_coords[0]
    already_selected[1] = max_coords[1]

    # minimum distance of all unselected samples to the two selected samples
    min_distances = np.min(ks_distance[:, max_coords], axis=1)
    for i in range(2, n_samples):
        # find the next sample with the largest minimum distance to any sample already selected
        already_selected[i] = np.argmax(min_distances)
        # get minimum distance of unselected samples to that sample only
        min_distances = np.minimum(min_distances, ks_distance[:, already_selected[i]])

    return already_selected



from datetime import date, datetime

import numpy as np

from astartes.samplers import AbstractSampler


class TimeBased(AbstractSampler):
    def _before_sample(self):
        # verify that the user provided time as the labels (i.e. args[2])
        if self.labels is None:
            msg = "Time based splitting requires the input labels to be a date or datetime object"
            raise ValueError(msg)

        # verify that labels (i.e. args[2]) contains the expected data type
        elif not all(isinstance(i, date) for i in self.labels) and not all(isinstance(i, datetime) for i in self.labels):
            msg = "Time based splitting requires the input labels to be an iterable of date or datetime objects"
            raise TypeError(msg)

    def _sample(self):
        """
        Implements the time-based sampler to create an extrapolation split.
        This places new data points in the testing set and older data points in the training set.
        """
        data = [(time, idx) for idx, time in zip(np.arange(len(self.labels)), self.labels)]
        sorted_list = sorted(data, reverse=False)

        self._samples_idxs = np.array([idx for time, idx in sorted_list], dtype=int)



"""
This sampler partitions the data based on the regression target y. It first sorts the
data by y value and then constructs the training set to have either the smallest (largest)
y values, the validation set to have the next smallest (largest) set of y values, and the
testing set to have the largest (smallest) y values.
"""

import numpy as np

from astartes.samplers import AbstractSampler


class TargetProperty(AbstractSampler):
    def _sample(self):
        """
        Implements the target property sampler to create an extrapolation split.
        """
        data = [(y, idx) for y, idx in zip(self.y, np.arange(len(self.y)))]

        # by default, the smallest property values are placed in the training set
        sorted_list = sorted(data, reverse=self.get_config("descending", False))

        self._samples_idxs = np.array([idx for time, idx in sorted_list], dtype=int)
